@@ -23,6 +23,8 @@
 # coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
 extensions = [
         "sphinx.ext.autodoc",
+        "sphinx.ext.intersphinx",
+        "sphinx.ext.mathjax",
         ]
 
 # Add any paths that contain templates here, relative to this directory.
@@ -43,7 +45,10 @@ copyright = '2010, Andreas Kloeckner'
 #
 # The short X.Y version.
 ver_dic = {}
-execfile("../pyvisfile/__init__.py", ver_dic)
+with open("../pyvisfile/__init__.py") as vpy_file:
+    version_py = vpy_file.read()
+
+exec(compile(version_py, "../pyvisfile/__init__.py", 'exec'), ver_dic)
 version = ".".join(str(x) for x in ver_dic["VERSION"])
 # The full version, including alpha/beta/rc tags.
 release = version
@@ -75,26 +80,23 @@ pygments_style = 'sphinx'
 # Options for HTML output
 # -----------------------
 
-try:
-    import sphinx_bootstrap_theme
-except:
-    from warnings import warn
-    warn("I would like to use the sphinx bootstrap theme, but can't find it.\n"
-            "'pip install sphinx_bootstrap_theme' to fix.")
-else:
-    # Activate the theme.
-    html_theme = 'bootstrap'
-    html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()
+html_theme = "alabaster"
 
-    # Theme options are theme-specific and customize the look and feel of a theme
-    # further.  For a list of options available for each theme, see the
-    # documentation.
-    html_theme_options = {
-            "navbar_fixed_top": "true",
-            "navbar_site_name": "Contents",
-            'bootstrap_version': '3',
-            'source_link_position': 'footer',
+html_theme_options = {
+        "extra_nav_links": {
+            "🚀 Github": "https://github.com/inducer/pyvisfile",
+            "💾 Download Releases": "https://pypi.python.org/pypi/pyvisfile",
             }
+        }
+
+html_sidebars = {
+    '**': [
+        'about.html',
+        'navigation.html',
+        'relations.html',
+        'searchbox.html',
+    ]
+}
 
 # The style sheet to use for HTML and HTML Help pages. A file of that name
 # must exist either in Sphinx' static/ path, or in one of the custom paths
@@ -145,7 +147,7 @@ htmlhelp_basename = 'PyVisfiledoc'
 
 # Grouping the document tree into LaTeX files. List of tuples
 # (source start file, target name, title, author, document class [howto/manual]).
-latex_documents = [("index", "pyvisfile.tex", 
+latex_documents = [("index", "pyvisfile.tex",
     "PyVisfile documentation", "Andreas Kloeckner",
     "manual")]
 
@@ -157,3 +159,9 @@ latex_documents = [("index", "pyvisfile.tex",
 
 # If false, no module index is generated.
 #latex_use_modindex = True
+
+intersphinx_mapping = {
+    'http://docs.python.org/': None,
+    'http://docs.scipy.org/doc/numpy/': None,
+    'https://documen.tician.de/modepy': None,
+}
